#include <node.h>
#include <nan.h>
#include <linux/fs.h>
#include <linux/spi/spidev.h>
#include <sys/ioctl.h>
#include <fcntl.h>

using namespace v8;
using namespace node;

extern "C" {
    #include <e9u_LSMD_defs.h>
    #include <e9u_LSMD.h>
}

static char Text_Buf[1024];
uv_mutex_t spiAccess;

class SPI_Transfer : public Nan::AsyncWorker {
    private:
        int fhwnd;
        int int_errno;

        uint32_t    spi_speed = 0;
        uint8_t     spi_mode = 0;
        uint8_t     spi_endian = 0;
        uint32_t    spi_read_len = 0;
        size_t      spi_buf_len = 0;
        uint8_t*    spi_tmp_buf = 0;

        uint8_t     wlen;
    
    public:
        ~SPI_Transfer() { free(spi_tmp_buf); } // we dont need to check if spi_tmp_buf == NULL (ISO-IEC 9899).

        SPI_Transfer(Local<Value> dev_path, uint32_t spi_speed, uint8_t spi_mode, uint8_t spi_endian, Local<Value> _writebuf, size_t spi_read_len, Nan::Callback *cb) :
            Nan::AsyncWorker(cb, "nan_spi:transfer"),
            fhwnd(open(*Nan::Utf8String(dev_path), O_RDWR)), 
            spi_speed(spi_speed),
            spi_mode(spi_mode),
            spi_endian(spi_endian), 
            spi_read_len(spi_read_len)
        {
            size_t writelen = 0;
            char* writedata = NULL;
            if (_writebuf->IsObject()) {
                Local<Object> writebuf =  Nan::To<Object>(_writebuf).ToLocalChecked();
                writelen = Buffer::Length(writebuf);
                writedata = Buffer::Data(writebuf);
            } 

            spi_buf_len = spi_read_len > writelen ? spi_read_len : writelen;
            spi_tmp_buf = (uint8_t*)malloc(spi_buf_len);

            if (writelen) 
                memcpy(spi_tmp_buf, writedata, writelen);

            if (spi_read_len > writelen) 
                memset(spi_tmp_buf+writelen, 0, spi_read_len-writelen);

            wlen = writelen;
        }

        void Execute () 
        {
            int ret = 0;

            uv_mutex_lock(&spiAccess);
            ret = ioctl(fhwnd, SPI_IOC_WR_MODE, &spi_mode);
            if (ret != -1) {
                ret = ioctl(fhwnd, SPI_IOC_WR_LSB_FIRST, &spi_endian);
                if (ret != -1) {
                    struct spi_ioc_transfer SPI_Msgs[2] = { 0 };

                    SPI_Msgs[0].tx_buf = (unsigned long)spi_tmp_buf;
                    SPI_Msgs[0].len = wlen;
                    SPI_Msgs[0].speed_hz = spi_speed;

                    SPI_Msgs[1].rx_buf = (unsigned long)spi_tmp_buf;
                    SPI_Msgs[1].len = spi_read_len;
                    SPI_Msgs[1].speed_hz = spi_speed;
                    
                    ret = ioctl(fhwnd, SPI_IOC_MESSAGE(2), SPI_Msgs);
                }
            }
            
            uv_mutex_unlock(&spiAccess);

            int_errno = (ret == -1) ? errno : 0;
        }
      
        void HandleOKCallback () {
            Nan::HandleScope();

            Local<Value> P_Error;
            Local<Value> P_Data;

            if (int_errno) 
            {
                snprintf(Text_Buf, sizeof(Text_Buf), "SPI error: %s (errno %i)", strerror(int_errno), int_errno);
                P_Error     = Nan::Error(Text_Buf);
                P_Data      = Nan::Null();
            }
            else
            {
                P_Error = Nan::Null();
                if (!spi_read_len) 
                {
                    P_Data = Nan::Null();
                }
                else
                {
                    P_Data = Nan::NewBuffer((char*)spi_tmp_buf, spi_read_len).ToLocalChecked();
                    spi_tmp_buf = NULL;
                }
            }

            Nan::TryCatch try_catch;
            Local<Value> Params[] = { P_Error, P_Data };
            callback->Call(2, Params, async_resource);

            if (try_catch.HasCaught()) {
                Nan::FatalException(try_catch);
            }
        };
};

NAN_METHOD(Transfer) {
    assert(info.Length() == 7);
    assert(info[0]->IsString()); // device_path
    assert(info[1]->IsNumber()); // spi_speed
    assert(info[2]->IsNumber()); // spi_mode
    assert(info[3]->IsNumber()); // spi_endian
    assert(info[4]->IsNull() || Buffer::HasInstance(info[4])); // writebuf
    assert(info[5]->IsNumber()); // spi_read_len
    assert(info[6]->IsFunction()); // callback

    Nan::AsyncQueueWorker(
        new SPI_Transfer(
            info[0], 
            Nan::To<uint32_t>(info[1]).FromJust(), 
            Nan::To<uint32_t>(info[2]).FromJust(), 
            Nan::To<uint32_t>(info[3]).FromJust(), 
            info[4], 
            Nan::To<uint32_t>(info[5]).FromJust(),
            new Nan::Callback(info[6].As<Function>())
        )
    );
}

NAN_METHOD(exec) {
    assert(info.Length() == 1);

    Nan::Utf8String nan_string(info[0]);
    std::string cc_string(*nan_string);
    const char* syscommand = cc_string.c_str();

    int ret = system(syscommand);


    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_read_camera (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_read_camera_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_read_camera(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_add_status_flag (unsigned int ui_camera_index, int i_flag);
NAN_METHOD(e9u_LSMD_add_status_flag_nan) {
    assert(info.Length() == 2);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t i_flag = Nan::To<uint32_t>(info[1]).FromJust();

    int ret =  e9u_LSMD_add_status_flag(ui_camera_index, i_flag);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_remove_status_flag (unsigned int ui_camera_index, int i_flag);
NAN_METHOD(e9u_LSMD_remove_status_flag_nan) {
    assert(info.Length() == 2);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t i_flag = Nan::To<uint32_t>(info[1]).FromJust();

    int ret =  e9u_LSMD_remove_status_flag(ui_camera_index, i_flag);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_get_dark_value (unsigned int ui_camera_index, unsigned int ui_channel, unsigned int ui_x_position, unsigned int ui_y_position);
NAN_METHOD(e9u_LSMD_get_dark_value_nan) {
    assert(info.Length() == 4);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_channel = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_x_position = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_y_position = Nan::To<uint32_t>(info[3]).FromJust();

    int ret =  e9u_LSMD_get_dark_value(ui_camera_index, ui_channel, ui_x_position, ui_y_position);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_get_pixel_value (unsigned int ui_camera_index, unsigned int ui_channel, unsigned int ui_x_position, unsigned int ui_y_position);
NAN_METHOD(e9u_LSMD_get_pixel_value_nan) {
    assert(info.Length() == 4);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_channel = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_x_position = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_y_position = Nan::To<uint32_t>(info[3]).FromJust();

    int ret =  e9u_LSMD_get_pixel_value(ui_camera_index, ui_channel, ui_x_position, ui_y_position);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_set_dark_value (unsigned int ui_camera_index, unsigned int ui_channel, unsigned int ui_x_position, unsigned int ui_y_position, unsigned int ui_value);
NAN_METHOD(e9u_LSMD_set_dark_value_nan) {
    assert(info.Length() == 5);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());
    assert(info[4]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_channel = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_x_position = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_y_position = Nan::To<uint32_t>(info[3]).FromJust();
    uint32_t ui_value = Nan::To<uint32_t>(info[4]).FromJust();

    int ret =  e9u_LSMD_set_dark_value(ui_camera_index, ui_channel, ui_x_position, ui_y_position, ui_value);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_set_pixel_value (unsigned int ui_camera_index, unsigned int ui_channel, unsigned int ui_x_position, unsigned int ui_y_position, unsigned int ui_value);
NAN_METHOD(e9u_LSMD_set_pixel_value_nan) {
    assert(info.Length() == 5);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());
    assert(info[4]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_channel = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_x_position = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_y_position = Nan::To<uint32_t>(info[3]).FromJust();
    uint32_t ui_value = Nan::To<uint32_t>(info[4]).FromJust();

    int ret =  e9u_LSMD_set_pixel_value(ui_camera_index, ui_channel, ui_x_position, ui_y_position, ui_value);

    info.GetReturnValue().Set(ret);
}

//DLL_API unsigned int DLL_CALL e9u_LSMD_get_frame_counter (unsigned int ui_camera_index, unsigned int ui_channel);
NAN_METHOD(e9u_LSMD_get_frame_counter_nan) {
    assert(info.Length() == 2);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_channel = Nan::To<uint32_t>(info[1]).FromJust();

    int ret =  e9u_LSMD_get_frame_counter(ui_camera_index, ui_channel);

    info.GetReturnValue().Set(ret);
}

//DLL_API unsigned int DLL_CALL e9u_LSMD_get_time_stamp (unsigned int ui_camera_index, unsigned int ui_channel, unsigned int ui_index);
/*
NAN_METHOD(e9u_LSMD_get_time_stamp_nan) {
    assert(info.Length() == 3);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_channel = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_index = Nan::To<uint32_t>(info[2]).FromJust();

    int ret =  e9u_LSMD_get_time_stamp(ui_camera_index, ui_channel, ui_index);

    info.GetReturnValue().Set(ret);
}
*/


//DLL_API int DLL_CALL e9u_LSMD_update_memory (unsigned int ui_camera_index, unsigned int ui_channel);
NAN_METHOD(e9u_LSMD_update_memory_nan) {
    assert(info.Length() == 2);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_channel = Nan::To<uint32_t>(info[1]).FromJust();

    int ret =  e9u_LSMD_update_memory(ui_camera_index, ui_channel);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_set_reg (unsigned int ui_camera_index, unsigned int ui_register, unsigned int ui_value);
NAN_METHOD(e9u_LSMD_set_reg_nan) {
    assert(info.Length() == 3);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_register = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_value = Nan::To<uint32_t>(info[2]).FromJust();

    int ret =  e9u_LSMD_set_reg(ui_camera_index, ui_register, ui_value);

    info.GetReturnValue().Set(ret);
}

//DLL_API unsigned int DLL_CALL e9u_LSMD_get_reg32 (unsigned int ui_camera_index, unsigned int ui_register, unsigned int ui_rx_tx);
NAN_METHOD(e9u_LSMD_get_reg32_nan) {
    assert(info.Length() == 3);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_register = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_rx_tx = Nan::To<uint32_t>(info[2]).FromJust();

    int ret =  e9u_LSMD_get_reg32(ui_camera_index, ui_register, ui_rx_tx);

    info.GetReturnValue().Set(ret);
}


//DLL_API unsigned int DLL_CALL e9u_LSMD_get_reg16 (unsigned int ui_camera_index, unsigned int ui_register, unsigned int ui_value_index, unsigned int ui_rx_tx);
NAN_METHOD(e9u_LSMD_get_reg16_nan) {
    assert(info.Length() == 4);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_register = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_value_index = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_rx_tx = Nan::To<uint32_t>(info[3]).FromJust();

    int ret =  e9u_LSMD_get_reg16(ui_camera_index, ui_register, ui_value_index, ui_rx_tx);

    info.GetReturnValue().Set(ret);
}

//DLL_API unsigned int DLL_CALL e9u_LSMD_get_reg8  (unsigned int ui_camera_index, unsigned int ui_register, unsigned int ui_value_index, unsigned int ui_rx_tx);
NAN_METHOD(e9u_LSMD_get_reg8_nan) {
    assert(info.Length() == 4);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_register = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_value_index = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_rx_tx = Nan::To<uint32_t>(info[3]).FromJust();

    int ret =  e9u_LSMD_get_reg8(ui_camera_index, ui_register, ui_value_index, ui_rx_tx);

    info.GetReturnValue().Set(ret);
}


//DLL_API int DLL_CALL e9u_LSMD_overwrite_reg32 (unsigned int ui_camera_index, unsigned int ui_register, unsigned int ui_value, unsigned int ui_rx_tx);
NAN_METHOD(e9u_LSMD_overwrite_reg32_nan) {
    assert(info.Length() == 4);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_register = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_value_index = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_rx_tx = Nan::To<uint32_t>(info[3]).FromJust();

    int ret =  e9u_LSMD_overwrite_reg32(ui_camera_index, ui_register, ui_value_index, ui_rx_tx);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_overwrite_reg16 (unsigned int ui_camera_index, unsigned int ui_register, unsigned int ui_value, unsigned int ui_value_index, unsigned int ui_rx_tx);
NAN_METHOD(e9u_LSMD_overwrite_reg16_nan) {
    assert(info.Length() == 5);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());
    assert(info[4]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_register = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_value = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_value_index = Nan::To<uint32_t>(info[3]).FromJust();
    uint32_t ui_rx_tx = Nan::To<uint32_t>(info[4]).FromJust();

    int ret =  e9u_LSMD_overwrite_reg16(ui_camera_index, ui_register, ui_value, ui_value_index, ui_rx_tx);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_overwrite_reg8  (unsigned int ui_camera_index, unsigned int ui_register, unsigned int ui_value, unsigned int ui_value_index, unsigned int ui_rx_tx);
NAN_METHOD(e9u_LSMD_overwrite_reg8_nan) {
    assert(info.Length() == 5);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());
    assert(info[4]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_register = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_value = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_value_index = Nan::To<uint32_t>(info[3]).FromJust();
    uint32_t ui_rx_tx = Nan::To<uint32_t>(info[4]).FromJust();

    int ret =  e9u_LSMD_overwrite_reg8(ui_camera_index, ui_register, ui_value, ui_value_index, ui_rx_tx);

    info.GetReturnValue().Set(ret);
}


//DLL_API int DLL_CALL e9u_LSMD_update_regs (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_update_regs_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_update_regs(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_begin (unsigned int ui_camera_index, const char *c_device);
NAN_METHOD(e9u_LSMD_begin_nan) {
    assert(info.Length() == 2);
    assert(info[0]->IsNumber());
    assert(info[1]->IsString());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    Nan::Utf8String nan_string(info[1]);
    std::string cc_string(*nan_string);
    const char* c_device = cc_string.c_str();

    int ret =  e9u_LSMD_begin(ui_camera_index, c_device);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_end (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_end_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_end(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_stop (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_stop_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_stop(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_board_info (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_board_info_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_board_info(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API char * DLL_CALL e9u_LSMD_board_type (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_board_type_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    char * ret =  e9u_LSMD_board_type(ui_camera_index);
 
    info.GetReturnValue().Set(String::NewFromUtf8(info.GetIsolate(), ret, v8::NewStringType::kNormal).ToLocalChecked());
}

//DLL_API double DLL_CALL e9u_LSMD_pixel_width (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_pixel_width_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    double ret =  e9u_LSMD_pixel_width(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API double DLL_CALL e9u_LSMD_pixel_height (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_pixel_height_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    double ret =  e9u_LSMD_pixel_height(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API unsigned int DLL_CALL e9u_LSMD_minimum_exposure (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_minimum_exposure_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_minimum_exposure(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API unsigned int DLL_CALL e9u_LSMD_step_exposure (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_step_exposure_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_step_exposure(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API unsigned int DLL_CALL e9u_LSMD_minimum_frame (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_minimum_frame_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_minimum_frame(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API unsigned int DLL_CALL e9u_LSMD_step_frame (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_step_frame_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_step_frame(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API unsigned int DLL_CALL e9u_LSMD_features (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_features_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_features(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_I2C_transfer (unsigned int ui_camera_index, unsigned int ui_i2c_adress, unsigned int ui_i2c_count, unsigned int ui_i2c_writeprotect, char *c_i2c_data);
NAN_METHOD(e9u_LSMD_I2C_transfer_nan) {
    assert(info.Length() == 5);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());
    assert(info[2]->IsNumber());
    assert(info[3]->IsNumber());
    assert(info[4]->IsString()); 

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_i2c_adress = Nan::To<uint32_t>(info[1]).FromJust();
    uint32_t ui_i2c_count = Nan::To<uint32_t>(info[2]).FromJust();
    uint32_t ui_i2c_writeprotect = Nan::To<uint32_t>(info[3]).FromJust();

    Nan::Utf8String nan_string(info[4]);
    std::string cc_string(*nan_string);
    const char* c_i2c_data = cc_string.c_str();

    int ret =  e9u_LSMD_I2C_transfer(ui_camera_index, ui_i2c_adress, ui_i2c_count, ui_i2c_writeprotect, (char*)c_i2c_data);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_eeprom_info (unsigned int ui_camera_index, unsigned int ui_pcb_number);
NAN_METHOD(e9u_LSMD_eeprom_info_nan) {
    assert(info.Length() == 2);
    assert(info[0]->IsNumber());
    assert(info[1]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    uint32_t ui_pcb_number = Nan::To<uint32_t>(info[1]).FromJust();

    int ret =  e9u_LSMD_eeprom_info(ui_camera_index, ui_pcb_number);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_begin_shm (unsigned int ui_camera_index, const char *c_device, const char *c_mmap_name);
NAN_METHOD(e9u_LSMD_begin_shm_nan) {
    assert(info.Length() == 3);
    assert(info[0]->IsNumber());
    assert(info[1]->IsString());
    assert(info[2]->IsString());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();
    const char* c_device = *Nan::Utf8String(info[1]);
    const char* c_mmap_name = *Nan::Utf8String(info[2]);

    int ret =  e9u_LSMD_begin_shm(ui_camera_index, c_device, c_mmap_name);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_end_shm (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_end_shm_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_end_shm(ui_camera_index);

    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_attach_shm (unsigned int ui_camera_index, const char *c_mmap_name);
NAN_METHOD(e9u_LSMD_attach_shm_nan) {
    assert(info.Length() == 2);
    assert(info[0]->IsNumber());
    assert(info[1]->IsString());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    Nan::Utf8String nan_string(info[1]);
    std::string cc_string(*nan_string);
    const char* c_mmap_name = cc_string.c_str();

    int ret =  e9u_LSMD_attach_shm(ui_camera_index, c_mmap_name);
    
    info.GetReturnValue().Set(ret);
}

//DLL_API int DLL_CALL e9u_LSMD_detach_shm (unsigned int ui_camera_index);
NAN_METHOD(e9u_LSMD_detach_shm_nan) {
    assert(info.Length() == 1);
    assert(info[0]->IsNumber());

    uint32_t ui_camera_index = Nan::To<uint32_t>(info[0]).FromJust();

    int ret =  e9u_LSMD_detach_shm(ui_camera_index);

    info.GetReturnValue().Set(ret);
}


NAN_MODULE_INIT(Initialize) {
    uv_mutex_init(&spiAccess);
    NAN_EXPORT(target, Transfer);
    NAN_EXPORT(target, exec);

    Export(target, "e9u_LSMD_read_camera", e9u_LSMD_read_camera_nan);
    Export(target, "e9u_LSMD_add_status_flag", e9u_LSMD_add_status_flag_nan);
    Export(target, "e9u_LSMD_remove_status_flag", e9u_LSMD_remove_status_flag_nan);
    Export(target, "e9u_LSMD_get_dark_value", e9u_LSMD_get_dark_value_nan);
    Export(target, "e9u_LSMD_get_pixel_value", e9u_LSMD_get_pixel_value_nan);
    Export(target, "e9u_LSMD_set_dark_value", e9u_LSMD_set_dark_value_nan);
    Export(target, "e9u_LSMD_set_pixel_value", e9u_LSMD_set_pixel_value_nan);
    Export(target, "e9u_LSMD_get_frame_counter", e9u_LSMD_get_frame_counter_nan);
    //Export(target, "e9u_LSMD_get_time_stamp", e9u_LSMD_get_time_stamp_nan);
    Export(target, "e9u_LSMD_update_memory", e9u_LSMD_update_memory_nan);
    Export(target, "e9u_LSMD_set_reg", e9u_LSMD_set_reg_nan);
    Export(target, "e9u_LSMD_get_reg32", e9u_LSMD_get_reg32_nan);
    Export(target, "e9u_LSMD_get_reg16", e9u_LSMD_get_reg16_nan);
    Export(target, "e9u_LSMD_get_reg8", e9u_LSMD_get_reg8_nan);
    Export(target, "e9u_LSMD_overwrite_reg32", e9u_LSMD_overwrite_reg32_nan);
    Export(target, "e9u_LSMD_overwrite_reg16", e9u_LSMD_overwrite_reg16_nan);
    Export(target, "e9u_LSMD_overwrite_reg8", e9u_LSMD_overwrite_reg8_nan);
    Export(target, "e9u_LSMD_update_regs", e9u_LSMD_update_regs_nan);
    Export(target, "e9u_LSMD_begin", e9u_LSMD_begin_nan);
    Export(target, "e9u_LSMD_end", e9u_LSMD_end_nan);
    Export(target, "e9u_LSMD_stop", e9u_LSMD_stop_nan);
    Export(target, "e9u_LSMD_board_info", e9u_LSMD_board_info_nan);
    Export(target, "e9u_LSMD_board_type", e9u_LSMD_board_type_nan);
    Export(target, "e9u_LSMD_pixel_width", e9u_LSMD_pixel_width_nan);
    Export(target, "e9u_LSMD_pixel_height", e9u_LSMD_pixel_height_nan);
    Export(target, "e9u_LSMD_minimum_exposure", e9u_LSMD_minimum_exposure_nan);
    Export(target, "e9u_LSMD_step_exposure", e9u_LSMD_step_exposure_nan);
    Export(target, "e9u_LSMD_minimum_frame", e9u_LSMD_minimum_frame_nan);
    Export(target, "e9u_LSMD_step_frame", e9u_LSMD_step_frame_nan);
    Export(target, "e9u_LSMD_features", e9u_LSMD_features_nan);
    Export(target, "e9u_LSMD_I2C_transfer", e9u_LSMD_I2C_transfer_nan);
    Export(target, "e9u_LSMD_eeprom_info", e9u_LSMD_eeprom_info_nan);
    Export(target, "e9u_LSMD_begin_shm", e9u_LSMD_begin_shm_nan);
    Export(target, "e9u_LSMD_end_shm", e9u_LSMD_end_shm_nan);
    Export(target, "e9u_LSMD_attach_shm", e9u_LSMD_attach_shm_nan);
    Export(target, "e9u_LSMD_detach_shm", e9u_LSMD_detach_shm_nan);
    
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_API_VERSION);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_CONTROL);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_EXP_TIME);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_FRAME_TIME);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_TRIG_DELAY);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_T_STAMP_EXP_STOP);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_T_STAMP_EXP_START);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_T_STAMP_TRIG);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_DARK_PRE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_PIXEL);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_DARK_POST);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_ADC);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_CH0_TYPE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_GLOBAL_CONTROL);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_VENDOR);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_BOARD_TYPE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_FPGA_VERSION);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_I2C);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_DATA_CH0);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_QUERY_REGS);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_STAY_ALIVE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_EXP_INT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_EXPOSURE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_TRIG_INT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_TRIG_EXT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_TRIG_USB);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_TRIG_DEBOUNCE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_PIX_DUMMY);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_PIX_DARK);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_PIX_VALUE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_FILL_DITH);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_FILL_ZERO);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_FILL_LIN);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_FILL_ONES);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e03_V_A);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e04_V_A);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e05_V_A);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e06_V_A);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e07_V_A);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e07_V_C);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e08_V_C);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e09_V_C);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e10_V_C);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e11_V_C);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e09_V_As);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e10_V_As);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e11_V_As);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e12_V_As);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_1e13_V_As);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_CHK_GAIN);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_DISCONNECT);
    NODE_DEFINE_CONSTANT(target, e9u_PDMD_C0C_DISCHARGE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_GC_CHANNEL_0);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_GC_STAY_ALIVE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_GC_SEND_CONF);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_GC_SEND_TIME);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_GC_SEND_TYPE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_C0C_DEFAULT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_GC_DEFAULT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_RX);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_TX);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_SH);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_DT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_VER);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_HOR);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_FPGA_BOARD);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_SENSOR_BOARD);

    //enum e9u_LSMD_return
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_FAIL);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_OK);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_ERROR_OPEN);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_ERROR_CLOSE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_ERROR_FLUSH);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_ERROR_WRITE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_TIMEOUT_WRITE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_ERROR_READ);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_TIMEOUT_READ);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_WARNING_UNKOWN);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_ERROR_PROTO);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_ERROR_MALLOC);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_ERROR_SEMAPHORE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_QUIT);

    //enum e9u_LSMD_status
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_NONE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_OPENED);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_QUERIED);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_VERIFIED);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_ALLOCATED);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_STARTED);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_SYNC);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_READ);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_COPY);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_LOCK);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_PAUSE);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_THREAD);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_REQUEST_QUIT);

    //enum e9u_LSMD_feature
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_EXP_INT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_EXP_EXT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_TRIG_INT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_TRIG_EXT);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_TRIG_USB);
    NODE_DEFINE_CONSTANT(target, e9u_LSMD_GPIO);
}

NODE_MODULE(Optosense, Initialize);